/**
  *  \file client/dialogs/chartconfig.cc
  *  \brief Starchart Configuration Dialog
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/chartconfig.h"
#include "client/markerconfig.h"
#include "client/widgets/chartdisplayconfig.h"
#include "client/widgets/chartlockconfig.h"
#include "client/widgets/colorselect.h"
#include "client/widgets/markerlist.h"
#include "client/widgets/markertype.h"
#include "cpluslib/holder.h"
#include "game/coord.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/cardtab.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/radiobutton.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "client/widgets/configstorage.h"
#include "game/lock.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"
#include "util/misc.h"
#include "ui/widgets/quit.h"

namespace {
    class WChartConfigDialog : public UIWindow {
     public:
        WChartConfigDialog();
        ~WChartConfigDialog();

     private:
        ObjectHolder h;
        WChartDisplayConfig* display_config;
        WMarkerList* marker_list;
        WChartLockConfig* lock_config;

        Observable<int> geo_kind;
        UIInputLine* geo_center;
        UIInputLine* geo_extent;

        WConfigStorageControl* display_storage;
        WConfigStorageControl* marker_storage;
        WConfigStorageControl* lock_storage;

        bool pconfig_warning;

        void init();
        void onOK();
        void onEditMarkerName();
        void onEditMarkerType();
        void onFocusChange();
        bool saveMapConfig();
    };
}

WChartConfigDialog::WChartConfigDialog()
    : UIWindow(_("Starchart Configuration"), BLUE_WINDOW, COLOR_GRAY),
      geo_kind(0),
      pconfig_warning(false)
{
    init();
}

WChartConfigDialog::~WChartConfigDialog()
{ }

void
WChartConfigDialog::init()
{
    // VBox
    //   UICardGroupTab
    //   UICardGroup
    //     VBox "Display"
    //       HBox
    //         UIFrameGroup
    //           WChartDisplayConfig
    //         UISpacer
    //       UISpacer
    //       WConfigStorageControl
    //     VBox "Geometry"
    //       UIStaticText "Map type"
    //       HBox
    //         UIStaticText <spacer> -- fixme?
    //         VBox
    //           UIRadioButton "Normal (plane)"
    //           UIRadioButton "Wrapped (rectangular, Sphere)"
    //       Grid
    //         "Center"  UIInputLine
    //         "Extent"  UIInputLine
    //       UISpacer
    //       WConfigStorageControl
    //     VBox "Markers"
    //       HBox
    //         WMarkerList
    //         VBox
    //           UIButton "Edit"
    //           UIButton "Name"
    //           UISpacer
    //       UISpacer
    //       WConfigStorageControl
    //     HBox "Mouse"
    //       UIFrameGroup
    //         WChartLockConfig
    //       UISpacer
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    //     UIButton "Help"
    setLayout(UIVBoxLayout::instance5);

    UICardGroup&   cards = h.add(new UICardGroup());
    UICardGroupTab& tabs = h.add(new UICardGroupTab(cards));

    // "Display" page
    UIGroup& display_group = h.add(new UIGroup(UIVBoxLayout::instance5, 1000));
    UIGroup& display_group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    display_storage = &h.add(new WConfigStorageControl(false));
    display_config = &h.add(new WChartDisplayConfig(2000));
    display_config->load();
    display_group1.add(UIFrameGroup::wrap(h, *display_config));
    display_group1.add(h.add(new UISpacer()));
    display_group.add(display_group1);
    display_group.add(h.add(new UISpacer()));
    display_group.add(*display_storage);
    for (int i = 0, n = getNumChartOptions(); i < n; ++i) {
        display_storage->addOption(getChartOption(i));
    }

    // "Geometry" page
    UIGroup& geo_group = h.add(new UIGroup(UIVBoxLayout::instance5, 1001));
    UIGroup& geo_group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& geo_group11 = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& geo_group2 = h.add(new UIGroup(h.add(new UIGridLayout(2))));
    geo_center = &h.add(new UIInputLine(20, 9, 'c', UIInputLine::ilf_Framed | UIInputLine::ilf_NoHi, 3003));
    geo_extent = &h.add(new UIInputLine(20, 9, 's', UIInputLine::ilf_Framed | UIInputLine::ilf_NoHi, 3004));

    switch (chart_conf.getMode()) {
     case GChartConfiguration::Flat:     geo_kind.set(0); break;
     case GChartConfiguration::Wrapped:  geo_kind.set(1); break;
     case GChartConfiguration::Circular: geo_kind.set(2); break;
    }
    GPoint geo_center_pt = chart_conf.getCenter();
    GPoint geo_extent_pt = chart_conf.getSize();
    geo_center->setText(format("%d,%d", geo_center_pt.x, geo_center_pt.y));
    if (geo_extent_pt.x == geo_extent_pt.y) {
        geo_extent->setText(format("%d", geo_extent_pt.x));
    } else {
        geo_extent->setText(format("%d,%d", geo_extent_pt.x, geo_extent_pt.y));
    }

    geo_group.add(h.add(new UIStaticText(_("Map type:"), UIStaticText::StaticTitle)));
    geo_group.add(geo_group1);
    geo_group1.add(h.add(new UIStaticText("  ", UIStaticText::StaticTitle)));
    geo_group1.add(geo_group11);
    geo_group11.add(h.add(new UIRadioButton(3000, _("Normal (plane)"), 'n', geo_kind, 0)));
    geo_group11.add(h.add(new UIRadioButton(3001, _("Wrapped (rectangular, Sphere)"), 'w', geo_kind, 1)));
    geo_group11.add(h.add(new UIRadioButton(3002, _("Round wrap (circular, PWrap)"), 'r', geo_kind, 2)));
    geo_group1.add(h.add(new UISpacer()));
    geo_group.add(geo_group2);
    geo_group2.add(h.add(new UIStaticText(_("Center:"), UIStaticText::StaticTitle)));
    geo_group2.add(*geo_center);
    geo_group2.add(h.add(new UIStaticText(_("Size:"), UIStaticText::StaticTitle)));
    geo_group2.add(*geo_extent);
    geo_group.add(h.add(new UISpacer()));
    geo_group.add(h.add(new UIFocusIterator(3000, 3004, UIFocusIterator::fi_Vert)));

    // "Markers" page
    UIGroup& marker_group = h.add(new UIGroup(UIVBoxLayout::instance5, 1002));
    UIGroup& marker_group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& marker_buttons = h.add(new UIGroup(UIVBoxLayout::instance5));
    marker_storage = &h.add(new WConfigStorageControl(false));
    marker_list = &h.add(new WMarkerList(4000));
    marker_group.add(marker_group1);
    marker_group1.add(*marker_list);
    marker_group1.add(marker_buttons);
    marker_group1.add(h.add(new UISpacer()));
    marker_group.add(h.add(new UISpacer()));
    marker_group.add(*marker_storage);
    for (int i = 0; i < ConfigMarkerOption::NUM_CANNED_MARKERS; ++i) {
        marker_storage->addOption(ConfigMarkerOption::getCannedMarker(i));
    }
    marker_buttons.add(h.add(new UIButton(4001, _("Space - Edit"), ' ')).addCall(this, &WChartConfigDialog::onEditMarkerType));
    marker_buttons.add(h.add(new UIButton(4002, _("N - Name"), 'n')).addCall(this, &WChartConfigDialog::onEditMarkerName));
    marker_buttons.add(h.add(new UISpacer()));

    // "Mouse" page
    UIGroup& mouse_group = h.add(new UIGroup(UIVBoxLayout::instance5, 1003));
    UIGroup& mouse_group1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    lock_storage = &h.add(new WConfigStorageControl(false));
    lock_config = &h.add(new WChartLockConfig(5000));
    lock_config->load();
    mouse_group1.add(UIFrameGroup::wrap(h, *lock_config));
    mouse_group1.add(h.add(new UISpacer()));
    mouse_group.add(mouse_group1);
    mouse_group.add(h.add(new UISpacer()));
    mouse_group.add(*lock_storage);
    for (int i = 0; i < lm_MAX; ++i) {
        lock_storage->addOption(getLockModeConfigOption(TLockMode(i)));
    }

    // Create pages
    cards.add(display_group);
    cards.add(geo_group);
    cards.add(marker_group);
    cards.add(mouse_group);
    tabs.addTab(1000, UIKeyString(_("Display")));
    tabs.addTab(1001, UIKeyString(_("Geometry")));
    tabs.addTab(1002, UIKeyString(_("Markers")));
    tabs.addTab(1003, UIKeyString(_("Mouse")));

    UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
    buttons.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WChartConfigDialog::onOK));
    buttons.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(_("Help"), 'h')));
    buttons.add(h.add(new WHelpWidget("pcc2:starchartopts")));

    add(tabs);
    add(cards);
    add(buttons);
    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Tab)));
    add(h.add(new UIQuit(0)));
    display_config->setFocusedWidget();
    pack();

    cards.sig_focus_changed.add(this, &WChartConfigDialog::onFocusChange);
}

void
WChartConfigDialog::onOK()
{
    // Map config can block saving the rest
    if (saveMapConfig()) {
        display_config->save();
        lock_config->save();
        marker_storage->save();
        display_storage->save();
        lock_storage->save();
        stop(1);
    } else {
        getWidgetById(1001)->setFocusedWidget();
    }
}

void
WChartConfigDialog::onEditMarkerName()
{
    ConfigMarkerOption& opt = ConfigMarkerOption::getCannedMarker(marker_list->getCurrent());

    // Edit name
    string_t note = opt.getNote();
    if (!UIInputLine::run(note, 1000, _("Edit Marker"), _("Name:"), 30, 0)) {
        return;
    }

    // Set it
    opt.setNote(note);
    opt.markUpdated(ConfigOption::srcUser);
    marker_list->drawWidget();
}

void
WChartConfigDialog::onEditMarkerType()
{
    ConfigMarkerOption& opt = ConfigMarkerOption::getCannedMarker(marker_list->getCurrent());

    // Edit type
    WMarkerTypeSelector mts;
    mts.setMarkerType(opt.getType());
    if (!mts.doStandardDialog(_("Edit Marker"))) {
        return;
    }

    // Edit color
    WColorSelector cs;
    cs.setColor(opt.getColorIndex());
    if (cs.doStandardDialog(_("Edit Marker"), false) != cs.rOK) {
        return;
    }

    // Set it
    opt.setType(mts.getMarkerType());
    opt.setColorIndex(cs.getColor());
    opt.markUpdated(ConfigOption::srcUser);
    marker_list->drawWidget();
}

void
WChartConfigDialog::onFocusChange()
{
    // Do we need to warn?
    // w can be null during destruction.
    UIBaseWidget* w = getWidgetById(1001);
    if (w != 0 && w->hasState(st_Focused) && !pconfig_warning) {
        pconfig_warning = true;

        // Check rules
        if (chart_conf.isSetFromGameConfig()) {
            // Redraw hack. The card group will redraw its content from its onFocusChange().
            // However, there is no guarantee that that has been called before this function,
            // and, with the current implementation if Signal/Slot, it isn't. Therefore, call
            // drawWidget() (which is what UICardGroup would do) before showing the message
            // box to make sure everything is drawn nicely.
            drawWidget();
            messageBox(_("These settings have been taking from the host configuration. "
                         "Changes will only last until PCC2 reads the configuration again. "
                         "Instead of manually changing Geometry settings, "
                         "it's usually better to work with a current copy of the host configuration."),
                       _("Starchart Configuration"));
        }
    }
}

/** Save map config. If map config is invalid, display a message box and return false.
    Derived from chartdlg.pas::CheckGeometryConf. */
bool
WChartConfigDialog::saveMapConfig()
{
    // Convert geometry selection
    GChartConfiguration::Mode m =
        geo_kind.get() == 1 ? GChartConfiguration::Wrapped
        : geo_kind.get() == 2 ? GChartConfiguration::Circular : GChartConfiguration::Flat;

    // Parse center
    GPoint center;
    if (!center.parseCoordinates(geo_center->getText())) {
        messageBox(_("The center coordinate specification is invalid. Please enter an "
                     "expression of the form \"2000,2000\" (X,Y)."),
                   _("Starchart Configuration"));
        return false;
    }

    // Parse size
    GPoint size;
    if (stringValue(geo_extent->getText(), size.x)) {
        size.y = size.x;
    } else {
        if (!size.parseCoordinates(geo_extent->getText())) {
            messageBox(_("The map extent specification is invalid. Please enter an "
                         "expression of the form \"1100,1100\" (sizeX,sizeY) or \"1100\" (size)."),
                       _("Starchart Configuration"));
            return false;
        }
    }

    // Validate ranges
    if (center.x < 500 || center.y < 500 || center.x > 4000 || center.y > 4000) {
        messageBox(_("The center coordinates must be between 500 to 4000."),
                   _("Starchart Configuration"));
        return false;
    }

    // Validate size for circular
    if (m == GChartConfiguration::Circular) {
        if (size.x != size.y) {
            messageBox(_("X and Y size must be equal for circular wrap."),
                       _("Starchart Configuration"));
            return false;
        }
        //    { I think this restriction is not actually needed. PCC can display negative
        //      positions just fine, it just cannot handle actually being there. }
        //    {IF (CenterX < SizeX + SPHERICAL_MAP_EXCESS) OR (CenterY < SizeX + SPHERICAL_MAP_EXCESS) THEN BEGIN
        //       IF (CenterX > SizeX) AND (CenterY > SizeY) THEN BEGIN
        //         NMessageBox('The center is too close to the origin. The outside part of the '+
        //                     'circular map would contain negative coordinates. If you need this '+
        //                     'map size, lower "Chart.Outside" in CONFIG2.CC.', Leerstring);
        //       END ELSE BEGIN
        //         NMessageBox('The center is too close to the origin. '+
        //                     'The map must not contain negative coordinates.', Leerstring);
        //       END;
        //       Exit;
        //     END;}
        //   END;
    }

    // Validate size
    if (size.x < 500 || size.y < 500 || size.x > 4000 || size.y > 4000 || size.x > center.x || size.y > center.y) {
        messageBox(_("The map size must be between 500 and 4000. "
                     "The values must be smaller than the center coordinate value "
                     "(the map must not contain negative coordinates)."),
                   _("Starchart Configuration"));
        return false;
    }

    // Is this actually a change?
    if (m != chart_conf.getMode() || size != chart_conf.getSize() || center != chart_conf.getCenter()) {
        // FIXME: must validate and reload planet XYs here
        //   IF NOT Memcmp(ChartConf, oldcc, Sizeof(ChartConf)) THEN BEGIN
        //     IF (PlayerId<>0) AND NOT LoadPlanetXY(CookieBitmap) THEN BEGIN
        //       NMessageBox('There are planets outside the size you specified. The minimum '+
        //                   'value for "Extent" for the current center coordinates is "' +
        //                    itoa(loadxy_minx) + ',' + itoa(loadxy_miny) + '".', Leerstring);
        //       ChartConf := oldcc;
        //       LoadPlanetXY(CookieBitmap);
        //       Exit;
        //     END;
        //   END;
        chart_conf.setConfig(m, center, size);
    }

    return true;
}

void
doChartConfig()
{
    WChartConfigDialog dlg;
    ui_root->centerWidget(dlg);
    ui_root->runModal(dlg);
}
