/**
  *  \file client/dialogs/navchart.cc
  *  \brief Ship Navigation Chart
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/navchart.h"
#include "client/dlg-visual.h"
#include "client/screens.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/navwidget.h"
#include "client/widgets/objlist.h"
#include "client/widgets/scanresult.h"
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/fleet.h"
#include "game/formula.h"
#include "game/objlist.h"
#include "game/ship.h"
#include "game/univ.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/prefix.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "ui/widgets/quit.h"

namespace {
    class WShipNavigationChartWithStop : public WShipNavigationChart {
     public:
        WShipNavigationChartWithStop(GUniverse& univ)
            : WShipNavigationChart(univ)
            { }

        void onDblClick()
            {
                stop(cm_Enter);
            }
    };
}

/** Choose chunnel mate from position.
    \param initiator Initiator of chunnel
    \param pt Location to check
    \return ship Id, 0 if none found or canceled */
static int
chooseChunnelMate(GUniverse& univ, const GShip& initiator, GPoint pt)
{
    GObjectList objs;
    for (int i = univ.ty_played_ships.findNextIndex(0); i != 0; i = univ.ty_played_ships.findNextIndex(i)) {
        GShip& sh = univ.ty_played_ships.getObjectByIndex(i);
        if (sh.getPos() == pt && isValidChunnelMate(initiator, sh)) {
            objs.addObject(GObjectReference(univ.ty_played_ships, i));
        }
    }

    if (objs.size() == 0) {
        messageBox(_("There are no potential chunnel mates at the current position."), _("Chunnel"));
        return 0;
    } else if (objs.size() == 1) {
        return objs[0]->getId();
    } else {
        GObjectReference id = WObjectList(0, 10, 20*font_ems[FONT_NORMAL], objs).doStandardDialog2(_("Chunnel"), "pcc2:navchart");
        if (id.isValid()) {
            return id->getId();
        } else {
            return 0;
        }
    }
}

static bool
handleConfirm(WShipNavigationChart& chart, GUniverse& univ, GShip& ship)
{
    if (chart.isChunnelMode()) {
        int fc = chooseChunnelMate(univ, ship, chart.getScannerTarget());
        if (fc == 0) {
            return false;
        }

        /* Clear speed and waypoint, set FC */
        setFleetWaypoint(univ, ship.getId(), ship.getPos());
        setFleetSpeed(univ, ship.getId(), 0);
        ship.setFCode(format("%03d", fc));

        /* Mate */
        GShip& mate = univ.getShip(fc);
        if (mate.isPlayable(GObject::Playable)) {
            /* For simplicity, use the fleet calls (although mates never are fleet members) */
            setFleetWaypoint(univ, fc, mate.getPos());
            setFleetSpeed(univ, fc, 0);
            // FIXME: if mate orbits a planet, PCC 1.x loads fuel onto it
        }

        /* Check whether chunnel works */
        GChunnelData cd;
        parseChunnelMission(ship, univ, cd);

        bool probFlag = false;
        string_t probStr = _("Please fix the following problems to make the chunnel work:");
        static const char*const problems[] = {
            N_("Initiator damaged"),
            N_("Initiator needs fuel"),
            N_("Initiator under tow"),
            N_("Mate damaged"),
            N_("Mate needs fuel"),
            N_("Mate moving"),
            N_("Mate under tow"),
        };
        static const int problemFlags[] = {
            GChunnelData::chf_Damaged,
            GChunnelData::chf_Fuel,
            GChunnelData::chf_Towed,
            GChunnelData::chf_MateDamaged,
            GChunnelData::chf_MateFuel,
            GChunnelData::chf_MateMoving,
            GChunnelData::chf_MateTowed,
        };
        for (uint_t i = 0; i < countof(problemFlags); ++i) {
            if (cd.failure & problemFlags[i]) {
                probStr += "\n" UTF_BULLET " ";
                probStr += _(problems[i]);
                probFlag = true;
            }
        }

        if (probFlag) {
            messageBox(probStr, _("Chunnel"));
        }
        return true;
    } else {
        setFleetWaypoint(univ, ship.getId(), chart.getScannerTarget());
        /* FIXME: setOptimumWarp does not belong here. Better do it in the script. But how? */
        setOptimumWarp(univ, ship.getId());
        if (ship.isFleetLeader()) {
            synchronizeFleet(univ, ship.getId());
        }
        return true;
    }
}

/** Display ship navigation chart. This will set the ship's waypoint
    and/or fcode.
    \param univ Universe we're in
    \param ship Ship to work on */
void
doShipNavigationChart(GUniverse& univ, GShip& ship)
{
    int size = ui_root->getExtent().h;

    // VBox
    //   UIFrameGroup
    //     WShipNavigationChart
    //   HBox
    //     WScanResult
    //     VBox
    //       UISpacer  -or-  HBox / Chunnel "C"
    //       HBox
    //         "OK"
    //         "ESC"
    //         "Help"

    UIWindow win(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    UIGroup& g2   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g22  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g222 = h.add(new UIGroup(UIHBoxLayout::instance5));

    WShipNavigationChartWithStop& chart = h.add(new WShipNavigationChartWithStop(univ));
    chart.setCenterPosition(ship.getPos());
    chart.setScannerOrigin(ship.getPos());
    chart.setScannerTarget(ship.getWaypoint());
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    chart.setCircleRadius(ship.isHyperdriving(),
                          square(ship.getWarp()) * (ship.canDoSpecial(hf_Gravitonic) ? 2 : 1));
    chart.sig_dblclick.add(&chart, &WShipNavigationChartWithStop::onDblClick);
    win.add(UIFrameGroup::wrap(h, chart));

    WScanResult& scanner = h.add(new WScanResult(false));
    scanner.setChart(chart);
    g2.add(scanner);
    g2.add(g22);

    /* Offer chunneling when ship is not member of a fleet only */
    if (ship.getFleetNumber() == 0
        && (ship.canDoSpecial(hf_FirecloudChunnel)
            || ship.canDoSpecial(hf_ChunnelSelf)
            || ship.canDoSpecial(hf_ChunnelOthers)))
    {
        UIGroup& g221 = h.add(new UIGroup(UIHBoxLayout::instance5));
        UIButton& btn = h.add(new UIButton(0, "C", 'c'));
        btn.setState(UIButton::bf_SmallSize, true);
        g221.add(h.add(new UISpacer()));
        g221.add(h.add(new UIStaticText(_("Make Chunnel"), UIStaticText::StaticNormal)));
        g221.add(btn);
        g22.add(g221);
        chart.setChunnelButton(&btn);
        chart.setChunnelShip(ship);
        btn.addCall(static_cast<WShipNavigationChart*>(&chart), &WShipNavigationChart::toggleChunnel);
    } else {
        g22.add(h.add(new UISpacer()));
    }
    g22.add(g222);

    g222.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    g222.add(h.add(new UIButton(cm_Escape, "ESC", SDLK_ESCAPE)).addStop());
    g222.add(h.add(new UIButton("H", 'h')));

    win.add(g2);
    win.add(h.add(new WHelpWidget("pcc2:navchart")));
    win.add(h.add(new UIPrefixArg()));
    win.add(h.add(new UIQuit(cm_Escape)));
    win.setExtent(GfxRect(0, 0, size - 50, size - 10));
    chart.setInitialZoom();

    ui_root->centerWidget(win);
    ui_root->add(win);
    while (ui_root->run() == cm_Enter) {
        if (handleConfirm(chart, univ, ship)) {
            break;
        }
    }
    ui_root->close(win);
}


/** Display generic navigation chart. This will allow the user to
    select a point, but is not bound to a specific ship. Therefore, it
    will not allow setting up chunnels.
    \param univ [in] Universe
    \param origin [in] Origin of selection (location of ship)
    \param target [in/out] Initially-selected waypoint
    \param hyp [in] true to show hyp-circles, false for regular warp circles
    \param r [in] regular warp circle radius, if any */
bool
doGenericNavigationChart(GUniverse& univ, GPoint origin, GPoint& target, bool hyp, int r)
{
    int size = ui_root->getExtent().h;

    // VBox
    //   UIFrameGroup
    //     WShipNavigationChart
    //   HBox
    //     WScanResult
    //     VBox
    //       UISpacer
    //       HBox
    //         "OK"
    //         "ESC"
    //         "Help"

    UIWindow win(_("Select Waypoint"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    UIGroup& g2   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g22  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g222 = h.add(new UIGroup(UIHBoxLayout::instance5));

    WShipNavigationChartWithStop& chart = h.add(new WShipNavigationChartWithStop(univ));
    chart.setCenterPosition(origin);
    chart.setScannerOrigin(origin);
    chart.setScannerTarget(target);
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    chart.setCircleRadius(hyp, r);
    chart.sig_dblclick.add(&chart, &WShipNavigationChartWithStop::onDblClick);
    win.add(UIFrameGroup::wrap(h, chart));

    WScanResult& scanner = h.add(new WScanResult(false));
    scanner.setChart(chart);
    g2.add(scanner);
    g2.add(g22);
    g22.add(h.add(new UISpacer()));
    g22.add(g222);

    g222.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    g222.add(h.add(new UIButton(cm_Escape, "ESC", SDLK_ESCAPE)).addStop());
    g222.add(h.add(new UIButton("H", 'h')));

    win.add(g2);
    win.add(h.add(new WHelpWidget("pcc2:navchart")));
    win.add(h.add(new UIPrefixArg()));
    win.add(h.add(new UIQuit(cm_Escape)));
    win.setExtent(GfxRect(0, 0, size - 50, size - 10));
    chart.setInitialZoom();

    ui_root->centerWidget(win);
    if (ui_root->runModal(win) == cm_Enter) {
        target = chart.getScannerTarget();
        return true;
    } else {
        return false;
    }
}
