/**
  *  \file game/unpack.cc
  *  \brief Unpacker core
  *
  *  FIXME: can we unify unpackShips / unpackPlanets / unpackBases
  *  to avoid code duplication? CCUnpack does that by manipulating
  *  binary representations (i.e. "planet has Id at offset 2"), but
  *  being a DOS program it doesn't have to deal with endianness
  *  issues etc.
  *
  *  FIXME: missing features compared to PCC 1.x's built-in unpacker:
  *  - receive util.dat file transfer
  *  - apply TRN file
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/unpack.h"
#include "cpluslib/format.h"
#include "game/checksum.h"
#include "game/global.h"
#include "game/rst.h"
#include "game/struct.h"
#include "io/dirs.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/greg.h"
#include "game/gamedirs.h"
#include "util/answer.h"
#include "util/progress.h"
#include "game/unpackmod.h"

/** Create default unpacker. */
GUnpacker::GUnpacker()
    : opt_format(WindowsFormat),
      opt_ignore_35(false),
      opt_target_ext(false),
      opt_errors(true),
      opt_force(false),
      opt_answer(&ConstantAnswerProvider::sayYes)
{ }

GUnpacker::~GUnpacker()
{ }

/** Main entry point. Unpacks the specified result file into the current game directory.
    \param result The result file
    \param player Player number
    \param pm Progress monitor
    \param modif Modifier that can intercept data
    \throws FileFormatException on various file format problems
    \throws FileProblemException on various other file problems */
void
GUnpacker::unpack(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif)
{
    Stream& s = result.getFile();

    // Clear memory
    target_ext_buffer.clear();

    // Start by reading GEN file to figure out player:
    TResultGen gen;
    s.seek(result.getSectionOffset(GResultFile::sec_Gen));
    getStructureT(s, gen);
    if (gen.player_id != player) {
        if (!opt_force) {
            throw FileFormatException(s, format(_("File is owned by player %d, should be %d"),
                                                gen.player_id, player));
        }
    }
    this->gen = GGen(gen);

    // Load control file. We don't care what kind it is, since we'll be rewriting it anyway.
    loadControl(*game_file_dir, player);

    // Unpack ships
    s.seek(result.getSectionOffset(GResultFile::sec_Ships));
    unpackShips(result, player, pm, modif);

    // Unpack planets
    s.seek(result.getSectionOffset(GResultFile::sec_Planets));
    unpackPlanets(result, player, pm, modif);

    // Unpack bases
    s.seek(result.getSectionOffset(GResultFile::sec_Bases));
    unpackBases(result, player, pm, modif);

    // Unpack targets
    s.seek(result.getSectionOffset(GResultFile::sec_Targets));
    unpackTargets(result, player, pm);

    // Unpack VCRs
    s.seek(result.getSectionOffset(GResultFile::sec_Vcrs));
    unpackVCRs(result, player, pm);

    // Unpack SHIPXY
    s.seek(result.getSectionOffset(GResultFile::sec_ShipXY));
    CTASSERT(GResultFile::sec_ShipXY + 1 == GResultFile::sec_Gen);
    unpackShipXY(result, player, result.getSectionOffset(GResultFile::sec_Gen) - result.getSectionOffset(GResultFile::sec_ShipXY), pm);

    // Unpack messages
    s.seek(result.getSectionOffset(GResultFile::sec_Messages));
    unpackMessages(result, player, pm);

    // Unpack Kore (minefields, racenames, storms, AVCs, Ufos)
    removeGameFile(format("kore%d.dat", player));
    if (!opt_ignore_35 && result.haveSection(GResultFile::sec_Kore)) {
        s.seek(result.getSectionOffset(GResultFile::sec_Kore));
        unpackKore(result, player, pm);
    }

    // Unpack Skore (more Ufos)
    removeGameFile(format("skore%d.dat", player));
    if (!opt_ignore_35 && result.haveSection(GResultFile::sec_Skore)) {
        s.seek(result.getSectionOffset(GResultFile::sec_Skore));
        unpackSkore(result, player, pm);
    }

    // FIXME: we don't unpack the LEECH file yet. Should we?
    // FIXME: we don't update the FIZZ file yet. Should we? PCC 1.x and CCUNPACK don't.

    // Create blank files (must be before finishDirectory)
    createBlankFiles(player);

    // Save target.ext file
    saveTargetExt(player);

    // Save gen file
    if (modif != 0) {
        modif->updateGen(this->gen);
        modif->finishDirectory();
    }
    saveGen(player);

    // Update indexes
    updateIndex(player);
    updateGameRegistry(this->gen.getTimestamp());

    // Save control file
    if (opt_format == WindowsFormat) {
        removeGameFile("control.dat");
        saveControl(*game_file_dir, player);
    } else {
        removeGameFile(format("contrl%d.dat", player));
        saveControl(*game_file_dir, 0);
    }
}

/** Unpack ship section. Creates a ship.dat/.dis pair. Also validates
    and updates the ship checksum in the gen block. */
void
GUnpacker::unpackShips(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif)
{
    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("ship%d.dat", player), Stream::C_WRITE | Stream::C_CREATE);
    Ptr<Stream> dis = game_file_dir->openFile(format("ship%d.dis", player), Stream::C_WRITE | Stream::C_CREATE);

    uint32 rst_checksum  = 0;
    uint32 game_checksum = 0;

    // Count
    TInt16 count;
    getStructureT(s, count);
    game_checksum += 2 * getChecksum(count.value);

    storeStructureT(*dat, count);
    storeStructureT(*dis, count);

    // Individual ships
    for (int32 i = 0; i < count.value; ++i) {
        // Read the ship
        char buffer[TShip::size];
        s.readT(buffer, sizeof(buffer));
        // Update checksum
        uint32 ship_checksum = getChecksum(buffer, sizeof(buffer));
        rst_checksum += ship_checksum;
        // Parse it
        TShip datship, disship;
        getStructure(buffer, datship);
        if (datship.ship_id <= 0 || datship.ship_id > NUM_SHIPS) {
            if (!opt_force)
                throw FileFormatException(s, format(_("Invalid ship Id %d"), datship.ship_id));
        } else {
            setControlEntry(ShipSection, datship.ship_id, ship_checksum);
        }
        // Fix it
        copyAndFix(datship, disship);
        if (modif != 0) {
            modif->updateShip(datship);
        }
        // Write it
        storeStructure(buffer, datship);
        game_checksum += getChecksum(buffer, sizeof(buffer));
        dat->writeT(buffer, sizeof(buffer));
        storeStructure(buffer, disship);
        dis->writeT(buffer, sizeof(buffer));
        game_checksum += getChecksum(buffer, sizeof(buffer));
        pm.tick();
    }

    // Verify checksum
    if (rst_checksum != gen.getChecksum(ShipSection))
        if (!opt_force)
            throw FileFormatException(s, _("Checksum mismatch in ship section"));

    // Write signatures
    dat->writeT(gen.getSig2(), 10);
    dis->writeT(gen.getSig1(), 10);
    game_checksum += getChecksum(gen.getSig2(), 10);
    game_checksum += getChecksum(gen.getSig1(), 10);
    gen.setChecksum(ShipSection, game_checksum);

    console.write(LOG_INFO, format(_("Unpacked %d ship%!1{s%}."), count.value));
}

/** Unpack planet section. Creates a pdata.dat/.dis pair. Also validates
    and updates the planet checksum in the gen block. */
void
GUnpacker::unpackPlanets(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif)
{
    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("pdata%d.dat", player), Stream::C_WRITE | Stream::C_CREATE);
    Ptr<Stream> dis = game_file_dir->openFile(format("pdata%d.dis", player), Stream::C_WRITE | Stream::C_CREATE);

    uint32 rst_checksum  = 0;
    uint32 game_checksum = 0;

    // Count
    TInt16 count;
    getStructureT(s, count);
    game_checksum += 2 * getChecksum(count.value);

    storeStructureT(*dat, count);
    storeStructureT(*dis, count);

    // Individual planets
    for (int32 i = 0; i < count.value; ++i) {
        // Read the planet
        char buffer[TPlanet::size];
        s.readT(buffer, sizeof(buffer));
        // Update checksum
        uint32 planet_checksum = getChecksum(buffer, sizeof(buffer));
        rst_checksum += planet_checksum;
        // Parse it
        TPlanet datplanet, displanet;
        getStructure(buffer, datplanet);
        if (datplanet.planet_id <= 0 || datplanet.planet_id > NUM_PLANETS) {
            if (!opt_force)
                throw FileFormatException(s, format(_("Invalid planet Id %d"), datplanet.planet_id));
        } else {
            setControlEntry(PlanetSection, datplanet.planet_id, planet_checksum);
        }
        // Fix it
        copyAndFix(datplanet, displanet);
        if (modif != 0) {
            modif->updatePlanet(datplanet);
        }
        // Write it
        storeStructure(buffer, datplanet);
        game_checksum += getChecksum(buffer, sizeof(buffer));
        dat->writeT(buffer, sizeof(buffer));
        storeStructure(buffer, displanet);
        game_checksum += getChecksum(buffer, sizeof(buffer));
        dis->writeT(buffer, sizeof(buffer));
        pm.tick();
    }

    // Verify checksum
    if (rst_checksum != gen.getChecksum(PlanetSection))
        if (!opt_force)
            throw FileFormatException(s, _("Checksum mismatch in planet section"));

    // Write signatures
    dat->writeT(gen.getSig2(), 10);
    dis->writeT(gen.getSig1(), 10);
    game_checksum += getChecksum(gen.getSig2(), 10);
    game_checksum += getChecksum(gen.getSig1(), 10);
    gen.setChecksum(PlanetSection, game_checksum);

    console.write(LOG_INFO, format(_("Unpacked %d planet%!1{s%}."), count.value));
}

/** Unpack starbase section. Creates a bdata.dat/.dis pair. Also validates
    and updates the base checksum in the gen block. */
void
GUnpacker::unpackBases(GResultFile& result, int player, ProgressMonitor& pm, GUnpackModifier* modif)
{
    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("bdata%d.dat", player), Stream::C_WRITE | Stream::C_CREATE);
    Ptr<Stream> dis = game_file_dir->openFile(format("bdata%d.dis", player), Stream::C_WRITE | Stream::C_CREATE);

    uint32 rst_checksum  = 0;
    uint32 game_checksum = 0;

    // Count
    TInt16 count;
    getStructureT(s, count);
    game_checksum += 2 * getChecksum(count.value);

    storeStructureT(*dat, count);
    storeStructureT(*dis, count);

    // Individual bases
    for (int32 i = 0; i < count.value; ++i) {
        // Read the base
        char buffer[TStarbase::size];
        s.readT(buffer, sizeof(buffer));
        // Update checksum
        uint32 base_checksum = getChecksum(buffer, sizeof(buffer));
        rst_checksum += base_checksum;
        // Parse it
        TStarbase datbase, disbase;
        getStructure(buffer, datbase);
        if (datbase.base_id <= 0 || datbase.base_id > NUM_PLANETS) {
            if (!opt_force)
                throw FileFormatException(s, format(_("Invalid starbase Id %d"), datbase.base_id));
        } else {
            setControlEntry(BaseSection, datbase.base_id, base_checksum);
        }
        // Fix it
        copyAndFix(datbase, disbase);
        if (modif != 0) {
            modif->updateBase(datbase);
        }
        // Write it
        storeStructure(buffer, datbase);
        game_checksum += getChecksum(buffer, sizeof(buffer));
        dat->writeT(buffer, sizeof(buffer));
        storeStructure(buffer, disbase);
        dis->writeT(buffer, sizeof(buffer));
        game_checksum += getChecksum(buffer, sizeof(buffer));
        pm.tick();
    }

    // Verify checksum
    if (rst_checksum != gen.getChecksum(BaseSection))
        if (!opt_force)
            throw FileFormatException(s, _("Checksum mismatch in starbase section"));

    // Write signatures
    dat->writeT(gen.getSig2(), 10);
    dis->writeT(gen.getSig1(), 10);
    game_checksum += getChecksum(gen.getSig2(), 10);
    game_checksum += getChecksum(gen.getSig1(), 10);
    gen.setChecksum(BaseSection, game_checksum);

    console.write(LOG_INFO, format(_("Unpacked %d starbase%!1{s%}."), count.value));
}

/** Unpack target section. Creates a target.dat file. If requested and
    necessary, stores excess targets in target_ext_buffer for later
    processing by saveTargetExt(). */
void
GUnpacker::unpackTargets(GResultFile& result, int player, ProgressMonitor& pm)
{
    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("target%d.dat", player), Stream::C_WRITE | Stream::C_CREATE);

    // Count
    TInt16 count;
    getStructureT(s, count);
    if (count.value < 0)
        throw FileFormatException(s, _("Target file is invalid"));

    // If target.ext is requested, only extract 50 targets; the others go in target.ext.
    int targetsInFile = count.value;
    if (opt_target_ext && targetsInFile > 50)
        targetsInFile = 50;

    // Save new count
    TInt16 newCount;
    newCount.value = targetsInFile;
    storeStructureT(*dat, newCount);

    // Copy to file
    dat->copyFrom(s, targetsInFile * TShipTarget::size, pm);

    // Copy remainder to target.ext buffer
    for (int i = targetsInFile; i < count.value; ++i) {
        char buffer[TShipTarget::size];
        s.readT(buffer, sizeof(buffer));
        target_ext_buffer.append(buffer, sizeof(buffer));
    }

    // Add signature
    dat->writeT(gen.getSig2(), 10);

    if (isVerbose() && targetsInFile != count.value)
        console.write(LOG_INFO, format(_("Unpacked %d visual contact%!1{s%} to TARGET%d.DAT."), targetsInFile, player));
    else
        console.write(LOG_INFO, format(_("Unpacked %d visual contact%!1{s%}."), count.value));
}

/** Unpack VCR section. Creates a vcr.dat file. */
void
GUnpacker::unpackVCRs(GResultFile& result, int player, ProgressMonitor& pm)
{
    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("vcr%d.dat", player), Stream::C_CREATE | Stream::C_WRITE);

    // Read count
    TInt16 count;
    getStructureT(s, count);
    if (count.value < 0)
        throw FileFormatException(s, _("VCR file is invalid"));
    storeStructureT(*dat, count);

    // Copy
    dat->copyFrom(s, count.value * TRawVcr::size, pm);

    // Add signature
    dat->writeT(gen.getSig2(), 10);

    console.write(LOG_INFO, format(_("Unpacked %d combat recording%!1{s%}."), count.value));
}

/** Unpack shipxy section. Creates a shipxy.dat file.
    \param result File to unpack
    \param player Player to unpack
    \param space  Space allocated for this file in the RST */
void
GUnpacker::unpackShipXY(GResultFile& result, int player, int32_t space, ProgressMonitor& pm)
{
    Stream& s = result.getFile();

    // Figure out size.
    // FIXME: should we warn if there's room for less than 500 ships?
    int32 size;
    if (space >= 999 * TShipXY::size)
        size = 999 * TShipXY::size;
    else
        size = 500 * TShipXY::size;

    Ptr<Stream> dat = game_file_dir->openFile(format("shipxy%d.dat", player), Stream::C_CREATE | Stream::C_WRITE);

    // Copy
    dat->copyFrom(s, size, pm);

    // Add signature
    dat->writeT(gen.getSig2(), 10);

    if (isVerbose())
        console.write(LOG_INFO, format(_("Unpacked ship location file with %d entries."), size / TShipXY::size));
}

/** Unpack messages. Creates mdata.dat file. */
void
GUnpacker::unpackMessages(GResultFile& result, int player, ProgressMonitor& pm)
{
    Stream& s = result.getFile();

    // Read count
    TInt16 count;
    getStructureT(s, count);
    if (count.value < 0)
        throw FileFormatException(s, _("Message file is invalid"));

    // Read index
    uint32 size = count.value * TIncomingMessageHeader::size;
    Array<char> dir(size);
    s.readT(dir.ptr(), size);

    // Validate index (FIXME: cut & paste from GInbox::loadInbox)
    int16 max_size = 0;
    for (int i = 0; i < count.value; ++i) {
        /* verify message directory entry */
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + TIncomingMessageHeader::size*i, hdr);
        if (hdr.address <= 0 || hdr.length <= 0)
            throw FileFormatException(s, _("Message file is invalid"));
        if (hdr.length > max_size)
            max_size = hdr.length;
    }

    // Now generate the file, Start by writing dummy count + invalid directory.
    Ptr<Stream> dat = game_file_dir->openFile(format("mdata%d.dat", player), Stream::C_CREATE | Stream::C_WRITE);

    dat->writeT("\0", 2);
    dat->writeT(dir.ptr(), size);

    // Copy each message
    Array<char> msgbuf(max_size);
    for (int i = 0; i < count.value; ++i) {
        // Read
        TIncomingMessageHeader hdr;
        getStructure(dir.ptr() + TIncomingMessageHeader::size*i, hdr);
        s.seek(hdr.address-1);
        s.readT(msgbuf.ptr(), hdr.length);

        // Write
        hdr.address = dat->getPos()+1;
        dat->writeT(msgbuf.ptr(), hdr.length);
        storeStructure(dir.ptr() + TIncomingMessageHeader::size*i, hdr);
        pm.tick();
    }

    // Write real header
    dat->seek(0);
    storeStructureT(*dat, count);
    dat->writeT(dir.ptr(), size);

    console.write(LOG_INFO, format(_("Unpacked %d message%!1{s%}."), count.value));
}

/** Unpack KORE file. Creates a kore.dat file and, if valid race names
    are included in the result, a race.nm file. */
void
GUnpacker::unpackKore(GResultFile& result, int player, ProgressMonitor& pm)
{
    using std::memcpy;
    using std::memcmp;

    // RST:                              File:
    //                                   2      turn number
    //                                   7      junk
    //                                   10     sig2
    //                                   83     junk
    //   500x8  minefields               500x8  minefields
    //   50x12  ion storms               50x12  ion storms
    //   50x4   bangs                    50x4   bangs
    //   682    race names
    //   100x78 Ufos                     100x78 Ufos
    //   4      "1211" or "1120"         4      "1211" or "1120"
    //                                   16     junk
    //   if 1120:                        if 1120:
    //     4    count                      4    count
    //     nx34 AVCs                       nx34 AVCs
    //                                   10     sig2

    Stream& s = result.getFile();

    Ptr<Stream> dat = game_file_dir->openFile(format("kore%d.dat", player), Stream::C_CREATE | Stream::C_WRITE);

    // Write the "junk" block
    char junk[102];
    zeroFill(junk);
    storeInt16(junk, gen.getTurnNumber());
    memcpy(junk+9, gen.getSig2(), 10);
    dat->writeT(junk, 102);

    // Copy minefields, ion storms, bangs
    dat->copyFrom(s, 500*8 + 50*12 + 50*4, pm);

    // Deal with race names. It seems this file is sometimes sent as blank, so avoid
    // damaging the user's file.
    char raceNameBuffer[682], origRaceNameBuffer[682];
    bool validRaceNames = false;
    s.readT(raceNameBuffer, sizeof(raceNameBuffer));
    for (uint32 i = 0; i < sizeof(raceNameBuffer) && !validRaceNames; ++i)
        if (uint8(raceNameBuffer[i]) > 32)
            validRaceNames = true;
    if (validRaceNames) {
        // Load old race names. Ignore any errors; those will mean that the existing
        // race.nm is truncated, unreadable or does not exist, which means we should
        // treat the received file as new.
        zeroFill(origRaceNameBuffer);
        try {
            Ptr<Stream> rn = spec_file_dir->openFile("race.nm", Stream::C_READ);
            rn->read(origRaceNameBuffer, sizeof(origRaceNameBuffer));
        }
        catch (...) {
        }

        // Only ask for replacement when the race names differ. This comparison will
        // also fail if we had an error above.
        if (memcmp(raceNameBuffer, origRaceNameBuffer, sizeof(raceNameBuffer)) != 0) {
            if (opt_answer->ask(qAcceptRaceNames, _("Race names have changed. Accept new names?"))) {
                Ptr<Stream> rn = game_file_dir->openFile("race.nm", Stream::C_CREATE | Stream::C_WRITE);
                rn->writeT(raceNameBuffer, sizeof(raceNameBuffer));
                console.write(LOG_INFO, _("Unpacked race name file."));
            }
        }
    }

    // Ufos
    dat->copyFrom(s, 78*100, pm);

    // Deal with AVCs. We have to do two things here: copy them to the
    // koreX.dat file, and duplicate the decrypted version to targetX.ext
    // if required. First, copy the marker and create the junk block.
    TInt32 count;
    getStructureT(s, count);
    storeStructureT(*dat, count);
    zeroFill(junk);
    dat->writeT(junk, 16);
    if (count.value == 0x30323131) {
        // AVCs present
        getStructureT(s, count);
        storeStructureT(*dat, count);
        if (count.value < 0 || count.value > NUM_SHIPS)
            throw FileFormatException(s, _("Unbelievable number of visual contacts"));

        for (int i = 0; i < count.value; ++i) {
            char buffer[TShipTarget::size];
            s.readT(buffer, sizeof(buffer));
            dat->writeT(buffer, sizeof(buffer));
            if (opt_target_ext) {
                // Target.ext requested, so decode and save it
                for (int n = 0; n < 20; ++n)
                    buffer[14+n] ^= (154-n);
                target_ext_buffer.append(buffer, sizeof(buffer));
            }
            pm.tick();
        }
    }

    // Add signature
    dat->writeT(gen.getSig2(), 10);

    if (isVerbose())
        console.write(LOG_INFO, _("Unpacked version 3.5 object file (kore)."));
}

/** Unpack SKORE file. Creates a skore.dat file, but only if it would not be empty. */
void
GUnpacker::unpackSkore(GResultFile& result, int player, ProgressMonitor& pm)
{
    // RST:                   File:
    //                        96   junk
    //                        5    "yAmsz" signature
    //   2    count           2    count
    //                        2    minor version ("01")
    //   nx78 Ufos            nx78 Ufos
    //                        if count <= 100: "no000"

    // We don't ever generate the "no000" case. In that case, we don't
    // generate a Skore file at all.

    Stream& s = result.getFile();

    TInt16 count;
    getStructureT(s, count);

    if (count.value <= 100) {
        if (isVerbose())
            console.write(LOG_INFO, _("Extended Ufo database exists but is empty."));
        return;
    }

    // File exists
    Ptr<Stream> dat = game_file_dir->openFile(format("skore%d.dat", player), Stream::C_CREATE | Stream::C_WRITE);

    // Generate the "junk" block.
    char junk[105];
    zeroFill(junk);
    memcpy(junk+96, "yAmsz", 5);
    storeInt16(junk+101, count.value);
    storeInt16(junk+103, result.getVersion());
    dat->writeT(junk, sizeof(junk));

    // Copy Ufo data
    dat->copyFrom(s, 78 * (count.value - 100), pm);

    console.write(LOG_INFO, format(_("Unpacked %d additional Ufo%!1{s%}."), count.value - 100));
}

/** Create blank files. Creates or erases files that must be blank at
    the start of the turn:
    - mess.dat/mess35.dat
    - cp.cc (used by ancient PCC)
    - cmd.txt (used by PCC, PCC2, VPA) */
void
GUnpacker::createBlankFiles(int player)
{
    static const char* const names[] = { "mess%d.dat", "mess35%d.dat" };

    // Create blank outbox file
    {
        Ptr<Stream> dat = game_file_dir->openFile(format(names[opt_format == WindowsFormat], player), Stream::C_CREATE | Stream::C_WRITE);
        dat->writeT("\0", 2);
    }

    // Erase other outbox file
    removeGameFile(format(names[opt_format != WindowsFormat], player));

    // Erase PCC additional command files
    removeGameFile(format("cp%d.cc", player));
    removeGameFile(format("cmd%d.txt", player));
}

/** Save GEN file. */
void
GUnpacker::saveGen(int player)
{
    Ptr<Stream> dat = game_file_dir->openFile(format("gen%d.dat", player), Stream::C_WRITE | Stream::C_CREATE);
    storeStructureT(*dat, gen.getGenRecord());
}

/** Save TARGET.EXT file. If some additional targets have accumulated
    in our buffer, write them out. */
void
GUnpacker::saveTargetExt(int player)
{
    if (target_ext_buffer.getSize() > 0) {
        // We have some additional targets
        Ptr<Stream> dat = game_file_dir->openFile(format("target%d.ext", player), Stream::C_WRITE | Stream::C_CREATE);

        // Write count
        TInt16 count;
        count.value = target_ext_buffer.getSize() / TShipTarget::size;
        storeStructureT(*dat, count);

        // Write data
        dat->writeT(target_ext_buffer.ptr(), target_ext_buffer.getSize());

        // Signature
        dat->writeT(gen.getSig2(), 10);

        if (isVerbose())
            console.write(LOG_INFO, format(_("Unpacked %d visual contact%!1{s%} to TARGET%d.EXT."), count.value, player));
    } else {
        // No additional targets, scrap the file
        removeGameFile(format("target%d.ext", player));
    }
}

/** Copy ship record and fix errors. Copies dat into dis. Fixes that
    can be transmitted to the host will be fixed only in dat, fixes
    that cannot be transmitted to the host will be fixed in both. */
void
GUnpacker::copyAndFix(TShip& dat, TShip& dis)
{
    // Fix errors that apply to both halves
    if (opt_errors) {
        // Dominate(?) creates ships with zero launchers of type X. Some clients
        // including ancient PCC versions erroneously treat these as torpers.
        if (dat.torp_launcher_count == 0)
            dat.torp_type = 0;
    }

    // Copy
    dis = dat;

    // Fix errors that apply to only one half. Maketurn will send those to the
    // host, fixing the problem in the host files as well.
    if (opt_errors) {
        // THost generates ships with negative warp when they have >100% damage.
        if (dat.warp < 0 || dat.warp > 10)
            dat.warp = 0;
        // Some programs generate nulls in string names. Convert to canonical format.
        storeBasicString(dat.name, getBasicString(dat.name));
        storeBasicString(dat.fcode, getBasicString(dat.fcode));
    }
}

/** Copy planet record and fix errors. Copies dat into dis. Fixes that
    can be transmitted to the host will be fixed only in dat, fixes
    that cannot be transmitted to the host will be fixed in both. */
void
GUnpacker::copyAndFix(TPlanet& dat, TPlanet& dis)
{
    // Copy
    dis = dat;

    // Fix errors
    if (opt_errors) {
        // Strings with embedded nulls
        storeBasicString(dat.fcode, getBasicString(dat.fcode));
    }
}

/** Copy starbase record and fix errors. Copies dat into dis. Fixes that
    can be transmitted to the host will be fixed only in dat, fixes
    that cannot be transmitted to the host will be fixed in both. */
void
GUnpacker::copyAndFix(TStarbase& dat, TStarbase& dis)
{
    // Copy
    dis = dat;
}

/** Update index. This updates the init.tmp file and marks our new
    player as present. */
void
GUnpacker::updateIndex(int player)
{
    char index[NUM_PLAYERS * 2];
    if (player <= 0 || player > NUM_PLAYERS)
        return;

    // Initialize
    zeroFill(index);

    // Load old index, if any. Deliberately ignore read errors.
    if (Ptr<Stream> f = game_file_dir->openFileNT("init.tmp", Stream::C_READ))
        f->read(index, sizeof(index));

    // Mark new player and write new index
    index[2*(player-1)]   = 1;
    index[2*(player-1)+1] = 0;

    Ptr<Stream> f = game_file_dir->openFile("init.tmp", Stream::C_WRITE | Stream::C_CREATE);
    f->writeT(index, sizeof(index));
}
