/**
  *  \file int/lex.h
  *  \brief Interpreter: Tokenizer
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_LEX_H
#define PCC_V2_INT_LEX_H

#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "game/error.h"

/** Tokenizer. This class can split a line into CCScript tokens. It roughly corresponds
    to PCC 1.1.17's tokenizer with the following exceptions:
    - unterminated strings are errors
    - identifiers can start with '_' or '$' (in PCC 1.x, they can only start with letters),
      but cannot end with dots.
    - 2147483647 is a valid integer constant

    Summarized, rules are:
    - whitespace separates tokens but is otherwise ignored
    - '%' starts a comment to end of line
    - decimal numeric literals are permitted, with decimal point, but no fancy stuff like
      hex or '1.0e+12'. Integer literals are automatically turned into floats if they leave
      32-bit range.
    - identifiers can contain letters, digits, '_', '$' and '.', but cannot start with a
      digit or '.', and cannot end with '.'. They are converted into upper-case. The special
      keywords 'TRUE' and 'FALSE' turn into boolean literals, 'PI' is a float literal, and
      'AND', 'OR', 'XOR', 'MOD', 'NOT' are operators.
    - strings can be delimited by apostrophes or double-quotes. Within a double-quoted
      string, the backslash can be used to quote the next character, to include a
      double-quote or backslash in a string.
    - other characters or character pairs are operator tokens, or invalid. */
class IntTokenizer {
 public:
    /// Token type.
    enum TokenType {
        // Specials:
        tEnd,                   // End of input
        tInvalid,               // Invalid character (character obtainable in String Value)

        // Literals:
        tInteger,               // Integer literal
        tFloat,                 // Float literal
        tString,                // String literal
        tBoolean,               // Boolean literal

        // Identifiers and Reserved Words:
        tIdentifier,            // Identifier
        tAND,
        tOR,
        tXOR,
        tNOT,
        tMOD,

        // Character Pairs:
        tNE,                    // <>
        tGE,                    // >=
        tLE,                    // <=
        tAssign,                // :=
        tArrow,                 // ->

        // Single-Character Tokesn:
        tAmpersand,
        tHash,
        tPlus,
        tMinus,
        tMultiply,
        tSlash,                 // real division
        tBackslash,             // integer division
        tCaret,                 // power
        tLParen,
        tRParen,
        tComma,
        tEQ,
        tLT,
        tGT,
        tColon,
        tSemicolon,
        tDot
    };

    IntTokenizer(const string_t& str);
    ~IntTokenizer();

    TokenType       getCurrentToken() const;
    int32_t         getCurrentInteger() const;
    const string_t& getCurrentString() const;
    double          getCurrentFloat() const;

    string_t        getRemainingLine() const;

    TokenType readNext();

    bool checkAdvance(TokenType t);
    bool checkAdvance(const char* keyword);

 private:
    string_t             line;
    string_t::size_type  pos;

    TokenType current_token;
    string_t  current_string;
    int32_t   current_int;
    double    current_float;

    void read();
    void readNumber();
};

/** Get type of current (most recently read) token. */
inline IntTokenizer::TokenType
IntTokenizer::getCurrentToken() const
{
    return current_token;
}

/** Get integer value of current (most recently read) token.
    This is the value associated with a tInteger or tBoolean token. */
inline int32_t
IntTokenizer::getCurrentInteger() const
{
    return current_int;
}

/** Get string value of current (most recently read) token.
    This is the value associated with a tString token, or the name of a tIdentifier token. */
inline const string_t&
IntTokenizer::getCurrentString() const
{
    return current_string;
}

/** Get float value of current (most recently read) token.
    This is the value associated with a tFloat token. */
inline double
IntTokenizer::getCurrentFloat() const
{
    return current_float;
}

/** Get remaining unparsed line. */
inline string_t
IntTokenizer::getRemainingLine() const
{
    return line.substr(pos);
}

/** Read next token.
    \return type of token read */
inline IntTokenizer::TokenType
IntTokenizer::readNext()
{
    read();
    return getCurrentToken();
}

#endif
