/**
  *  \file io/streams.cc
  *  \brief Stream Implementation
  *
  *  The stream base class, and the simple standard stream classes.
  *  Implementation of the basic stream primitives and the SDL stream
  *  bridge.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cerrno>
#include "io/streams.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "util/nls.h"
#include "util/progress.h"

/*! \class Stream

    This class provides an abstract interface to streams, mostly
    intended for binary I/O. */

Stream::Stream() throw()
    : first_child(0), active_child(0)
{ }

/** \fn Stream::read(char*,std::size_t)
    \brief Read data

    Should read up to max bytes into the buffer; returns the number of
    bytes read. Returns zero if and only if EOF is reached. Need only
    be supported if capabilities include C_READ.

    \throws FileIOException on error. If appropriate, the stream may
    also go into an error state (where future reads return zero). */

/** \fn Stream::write(const char*,std::size_t)
    \brief Write data

    Should write up to max bytes from the buffer to the stream;
    returns the number of bytes written. Should return less than specified
    in max only if output device is full. Need only be supported if
    capabilities include C_WRITE.

    \throws FileIOException on error. If appropriate, the stream may
    also go into an error state (where future writes return zero). */

/** \fn Stream::seek(fpos_t)
    \brief Set file pointer

    Should position the file pointer at the specified position. The first
    byte in the stream is position 0. Need only be supported if
    capabilities include C_SEEK.

    \throws FileIOException on error */

/** \fn Stream::getPos()
    \brief Get file pointer

    Should return the current position of the file pointer. Need only
    be supported if capabilities include C_SEEK.

    \returns negative value on error. */

/** \fn Stream::getSize()
    \brief Get data size

    Should return the total number of bytes that can be read from the
    stream. Need only be supported if capabilities include C_SEEK.

    \returns negative value on error. */

/** \fn Stream::getCapabilities() const
    \brief Get capabilities of this stream

    Should return the capabilities of this stream, as a bitfield of the
    C_xxx constants. */

Stream::~Stream()
{
    while (first_child)
        first_child->disconnect();
    ASSERT(active_child == 0);
}

/** Stream is not readable. Call from your read() implementation from
    an unreadable stream. */
std::size_t
Stream::notReadable()
{
    throw FileProblemException(*this, _("Reading not supported"));
}

/** Stream is not writable. Call from your write() implementation from
    an unwritable stream. */
std::size_t
Stream::notWritable()
{
    throw FileProblemException(*this, _("Writing not supported"));
}

/** Stream is not seekable. Call from your seek() implementation. */
void
Stream::notSeekable()
{
    throw FileProblemException(*this, _("Seeking not supported"));
}

/** Invalid file position. Call from your seek() implementation. */
void
Stream::invalidFilePosition()
{
    throw FileProblemException(*this, _("Invalid file position"));
}


/** Check stream capabilities.
    \param what Capabilities to check for
    \return true iff the stream can do all the required things */
bool
Stream::canDo(int what)
{
    return (getCapabilities() & what) == what;
}

/** Read stream into buffer. Reads the complete stream (starting at
    the current file position) and appends it to the specified buffer.
    \param buffer [out] Buffer */
void
Stream::readBuffer(Buffer<char>& buffer)
{
    char b[4096];

    while(std::size_t n = read(b, sizeof b))
        buffer.append(b, n);
}

/** Read data, throw on error. Like read(), but throws FileTooShortException
    if not enough data can be read.
    \param buffer Buffer to read into
    \param count Number of bytes to read */
void
Stream::readT(char* buffer, std::size_t count)
{
    if (read(buffer, count) != count)
        throw FileTooShortException(*this);
}

/** Write data, throw on error. Like write(), but throws FileWriteException
    if not enough data can be read.
    \param buffer Buffer to write into
    \param count Number of bytes to write */
void
Stream::writeT(const char* buffer, std::size_t count)
{
    if (write(buffer, count) != count)
        throw FileWriteException(*this);
}

/** Copy from another stream. Copies as much as it gets. */
void
Stream::copyFrom(Stream& other, ProgressMonitor& pm)
{
    char buffer[4096];
    while (std::size_t n = other.read(buffer, sizeof(buffer))) {
        writeT(buffer, n);
        pm.tick();
    }
}

/** Copy from another stream. Copies exactly \c size bytes.
    Throws when the source stream does not have that many bytes. */
void
Stream::copyFrom(Stream& other, uint32_t size, ProgressMonitor& pm)
{
    char buffer[4096];
    while (size > 0) {
        uint32_t now = size > sizeof(buffer) ? sizeof(buffer) : size;
        other.readT(buffer, now);
        writeT(buffer, now);
        size -= now;
        pm.tick();
    }
}

/** Stream factory. \overload
    \returns stream reading file \c name opened with mode \c mode
    \throw FileOpenException when open fails */
Ptr<Stream>
Stream::openFile(const char* name, int mode)
{
    return openFile(string_t(name), mode);
}

/** Stream factory. Like openFile(), but instead of throwing an
    exception, this one returns 0 if the file can't be opened. */
Ptr<Stream>
Stream::openFileNT(const string_t& name, int mode)
{
    try {
        return openFile(name, mode);
    }
    catch(...) {
        return 0;
    }
}

/** Get separate handle ("duplicate") to this stream. In case
    simultaneous read/write for this stream is required, you can
    create any number of duplicates. All these duplicates are
    equivalent in function. They all have their own file pointer (and
    other state, in case such is introduced later). The following
    restrictions apply to the "master stream", i.e. the one which you
    call dup() on:

    - closing (deleting) the master stream may invalidate the
      duplicates.
    - the master stream is not protected from the duplicates, i.e.
      changing the master state may change a duplicate's state.
      If you need access to a master stream after you gave out a
      duplicate, you must use an own duplicate, too.

    \default Returns a ChildStream. Derived classes may override this
    if they have a more efficient method. They may also relax the
    above restrictions. */
Ptr<Stream>
Stream::dup()
{
    return new ChildStream(*this);
}

/***************************** SubStreamBase *****************************/

/*! \class SubStreamBase
    This class provides a "window" into a stream, by restricting certain
    access.

    setRange() sets origin and length:
    - origin >= 0: users can only access positions after that one.
    - origin < 0 disables seeking, so users can only read forward.
    - length >= 0: users can only access this many bytes after origin.
    - length < 0: length is not limited.

    setDisallowedOperations() sets a bitfield of disallowed access modes
    (C_READ to disable reading etc.).

    The SubStreamBase class is intended as a base class for things that
    need this functionality. Users usually instantiate the SubStream class
    instead. */
SubStreamBase::SubStreamBase(Stream& worker)
    : origin(-1), length(-1), position(0), worker(worker), op_mask(0)
{ }

/** Set ranges. See class description for explanation. */
void
SubStreamBase::setRange(fpos_t aorigin, fpos_t alength)
{
    origin   = aorigin;
    length   = alength;
    position = 0;
    seek(0);
}

int
SubStreamBase::getCapabilities() const
{
    int op = worker.getCapabilities() & ~op_mask;
    if (origin < 0)
        op &= C_SEEK;
    return op;
}

std::size_t
SubStreamBase::read(char* buffer, std::size_t max)
{
    if (op_mask & C_READ)
        return notReadable();

    if(length >= 0 && max > length - position)
        max = length - position;

    std::size_t r = worker.read(buffer, max);
    position += r;
    return r;
}

std::size_t
SubStreamBase::write(const char* buffer, std::size_t max)
{
    if(op_mask & C_WRITE)
        return notWritable();
    if(length >= 0 && max > length - position)
        max = length - position;

    std::size_t w = worker.write(buffer, max);
    position += w;
    return w;
}

void
SubStreamBase::seek(fpos_t pos)
{
    if ((op_mask & C_SEEK) || origin < 0)
        notSeekable();
    if (pos < 0 || (length >= 0 && pos > length))
        invalidFilePosition();
    worker.seek(pos + origin);
    position = pos;
}

SubStreamBase::fpos_t
SubStreamBase::getPos()
{
    return position;
}

SubStreamBase::fpos_t
SubStreamBase::getSize()
{
    fpos_t o = worker.getSize();
    if (o < 0)
        return length;
    else if (origin >= 0 && origin <= o)
        return std::min(fpos_t(o - origin), length);
    else
        return -1;
}

string_t
SubStreamBase::getName()
{
    return worker.getName();
}

/******************************* SubStream *******************************/

/** \class SubStream
    This is the user-callable version of SubStreamBase (see there for
    more information). */

SubStream::SubStream(Stream& worker)
    : SubStreamBase(worker)
{ }

/****************************** ChildStream ******************************/

/*! \class ChildStream
    \brief Alternative Handle to a Stream

    This class provides an alternate handle to a stream. It has
    independant status, i.e. two ChildStream's created from the same
    Stream will behave as two completely independant streams, a status
    change on one will not affect the other.

    Generally, this class is used indirectly via Stream::dup(), it is
    not used directly.

    \see Stream::dup() */

/** Constructor. Create a duplicate for /parent/. */
ChildStream::ChildStream(Stream& parent)
    : parent(&parent),
      next_child(parent.first_child),
      prev_child(&parent.first_child),
      current_pos(parent.getPos())

{
    if (next_child) {
        next_child->prev_child = &next_child;
    }
    parent.first_child = this;
}

/** Constructor. Create a disfunctional stream. This one will not
    react onto any call. */
ChildStream::ChildStream() throw()
    : parent(0), next_child(0), prev_child(0), current_pos(0)
{ }

ChildStream::~ChildStream()
{
    disconnect();
}

/** Activate this child. Restores the parent's state in case we were
    inactive. */
bool
ChildStream::activate()
{
    if (!parent)
        return false;
    if (parent->active_child == this)
        return true;
    if (parent->active_child)
        parent->active_child->deactivate();
    parent->seek(current_pos);
    parent->active_child = this;
    return true;
}

/** Deactivate this child. Saves our state for later restore. */
void
ChildStream::deactivate()
{
    if (parent && parent->active_child == this) {
        current_pos = parent->getPos();
        parent->active_child = 0;
    }
}

/** Disconnect this child from its parent. Object will be unusable
    afterwards. */
void
ChildStream::disconnect() throw()
{
    if (parent) {
        ASSERT(prev_child);
        ASSERT(*prev_child == this);
        *prev_child = next_child;
        if (next_child) {
            next_child->prev_child = prev_child;
        }
        if (parent && parent->active_child == this) {
            parent->active_child = 0;
        }
        parent = 0;
        next_child = 0;
        prev_child = 0;
    }
}

/** Check whether this child is active. */
bool
ChildStream::isActive() const throw()
{
    return parent && parent->active_child == this;
}

std::size_t
ChildStream::read(char* buffer, std::size_t max)
{
    if (activate())
        return parent->read(buffer, max);
    else
        return notReadable();
}

std::size_t
ChildStream::write(const char* buffer, std::size_t max)
{
    if (activate())
        return parent->write(buffer, max);
    else
        return notWritable();
}

int
ChildStream::getCapabilities() const
{
    if (parent)
        return parent->getCapabilities();
    else
        return 0;
}

void
ChildStream::seek(fpos_t pos)
{
    if (activate()) {
        parent->seek(pos);
    } else {
        notSeekable();
    }
}

Stream::fpos_t
ChildStream::getPos()
{
    if (isActive())
        return parent->getPos();
    else
        return current_pos;
}

Stream::fpos_t
ChildStream::getSize()
{
    if (activate())
        return parent->getSize();
    else
        return -1;
}

string_t
ChildStream::getName()
{
    if (activate())
        return parent->getName();
    else
        return "?";
}

Ptr<Stream>
ChildStream::dup()
{
    if (parent)
        return new ChildStream(*parent);
    else
        return new ChildStream();
}
