/**
  *  \file reshack/palette.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "reshack/palette.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/decimalsel.h"
#include "reshack/colors.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "util/nls.h"

class RHPaletteEditorWindow::Preview : public UIWidget {
 public:
    Preview()
        { }

    void setColor(uint32_t color);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);

 private:
    uint32_t color;
};

void
RHPaletteEditorWindow::Preview::setColor(uint32_t color)
{
    this->color = color;
    drawWidget();
}

void
RHPaletteEditorWindow::Preview::drawContent(GfxCanvas& can)
{
    drawSolidBar(can, getExtent(), color);
    can.update(getExtent());
}

void
RHPaletteEditorWindow::Preview::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(100, 30);
    info.flags = li_FlexibleH;
}


RHPaletteEditorWindow::RHPaletteEditorWindow(Ptr<GfxPixmap> pixmap, int colorIndex)
    : UIWindow(_("Palette"), BLUE_WINDOW, COLOR_GRAY),
      pixmap(pixmap),
      colorIndex(colorIndex),
      preview(0),
      sel_red(0),
      sel_green(0),
      sel_blue(0)
{
    init();
}

RHPaletteEditorWindow::~RHPaletteEditorWindow()
{ }

void
RHPaletteEditorWindow::init()
{
    // Get original color
    if (colorIndex < pixmap->getSystemSurface()->format->palette->ncolors) {
        orig_color = pixmap->getSystemSurface()->format->palette->colors[colorIndex];
    } else {
        std::memset(&orig_color, 0, sizeof(orig_color));
    }

    // Build window
    //   VBox
    //     Grid
    //       "Red"   UINumberSelector
    //       "Green" UINumberSelector
    //       "Blue"  UINumberSelector
    //     Preview
    //     HBox
    //       UIButton "OK"
    //       UIButton "Cancel"
    //       UISpacer
    setLayout(UIVBoxLayout::instance5);

    // Component editors
    sel_red = &h.add(new UIDecimalSelector(100, 0, 255, 10));
    sel_red->setValue(orig_color.r);
    sel_red->sig_changed.add(this, &RHPaletteEditorWindow::onChange);
    sel_green = &h.add(new UIDecimalSelector(101, 0, 255, 10));
    sel_green->setValue(orig_color.g);
    sel_green->sig_changed.add(this, &RHPaletteEditorWindow::onChange);
    sel_blue = &h.add(new UIDecimalSelector(102, 0, 255, 10));
    sel_blue->setValue(orig_color.b);
    sel_blue->sig_changed.add(this, &RHPaletteEditorWindow::onChange);

    // Components
    UIGroup& comp_group = h.add(new UIGroup(h.add(new UIGridLayout(2))));
    comp_group.add(h.add(new UIStaticText(_("Red"), UIStaticText::StaticTitle)));
    comp_group.add(*sel_red);
    comp_group.add(h.add(new UIStaticText(_("Green"), UIStaticText::StaticTitle)));
    comp_group.add(*sel_green);
    comp_group.add(h.add(new UIStaticText(_("Blue"), UIStaticText::StaticTitle)));
    comp_group.add(*sel_blue);
    add(comp_group);

    // Preview
    preview = &h.add(new Preview());
    preview->setColor(colorIndexToScreenColor(pixmap, colorIndex));
    add(*preview);

    // Buttons
    UIGroup& button_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    button_group.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &RHPaletteEditorWindow::onOK));
    button_group.add(h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE)).addCall(this, &RHPaletteEditorWindow::onCancel));
    button_group.add(h.add(new UISpacer()));
    add(button_group);

    add(h.add(new UIFocusIterator(100, 102, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab | UIFocusIterator::fi_Wrap)));

    pack();
}

void
RHPaletteEditorWindow::onChange()
{
    SDL_Color c;
    c.r = sel_red->getValue();
    c.g = sel_green->getValue();
    c.b = sel_blue->getValue();
    setPalette(c);
    preview->setColor(colorIndexToScreenColor(pixmap, colorIndex));
}

void
RHPaletteEditorWindow::onOK()
{
    stop(1);
}

void
RHPaletteEditorWindow::onCancel()
{
    setPalette(orig_color);
    stop(0);
}

void
RHPaletteEditorWindow::setPalette(SDL_Color& c)
{
    SDL_SetPalette(pixmap->getSystemSurface(), SDL_LOGPAL | SDL_PHYSPAL, &c, colorIndex, 1);
}
