/**
  *  \file misc.h
  *  \brief Miscellaneous utilities
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_MISC_H
#define PCC_V2_MISC_H

#include <algorithm>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "defs.h"
#include "cpluslib/smallset.h"

bool        stringMatch(const char* pattern, const char* tester);
inline bool stringMatch(const char* pattern, const string_t& tester)
{ return stringMatch(pattern, tester.c_str()); }

bool stringValue(string_t s, long& val);
bool stringValue(string_t s, long& val, string_t::size_type& pos);
bool stringValue(string_t s, int& val);
bool stringValue(string_t s, int& val, string_t::size_type& pos);

bool parseFlags(const string_t& s, const char* data, int* value,
                SmallSet<int>* races, bool accept_bad_flags = false);
bool parseRange(const string_t& s, int& min, int& max, string_t::size_type& pos);

string_t  numToString(int32 number);
string_t  clansToString(int32 number);
string_t  numToRoman(int32 number);

#ifndef HAVE_COPY_N
/* This one is a SGI extension. */
template <class InputIterator, class Size, class OutputIterator>
inline void
copy_n_internal(InputIterator first, Size count, OutputIterator result,
                std::input_iterator_tag)
{
    while (count > 0) {
        *result = *first;
        ++first;
        ++result;
        --count;
    }
}

template <class InputIterator, class Size, class OutputIterator>
inline void
copy_n_internal(InputIterator first, Size count, OutputIterator result,
                std::random_access_iterator_tag)
{
    std::copy(first, first + count, result);
}

/** Copy sequence of given length.
    \param first start of sequence
    \param count number of elements
    \param result output sequene

    \note According to the manpage, this function returns OutputIterator.
    In reality it returns pair<InputIterator,OutputIterator>. So we cannot
    safely use the result anyway :-/ */
template <class InputIterator, class Size, class OutputIterator>
inline void
copy_n(InputIterator first, Size count, OutputIterator result)
{
    typedef typename std::iterator_traits<InputIterator>::iterator_category ic;
    copy_n_internal(first, count, result,
                    ic());
}
#else
using std::copy_n;
#endif

/** Safe append to sequence-of-pointers. <tt>safeAppend(c, new Ele())</tt>
    is identical to <tt>c.push_back(new Ele())</tt> but avoids
    the memory leak if push_back throws.
    \return p */
template<class Container, class Ele>
Ele*
safeAppend(Container& c, Ele* p)
{
    try {
        c.push_back(p);
        return p;
    }
    catch (...) {
        delete p;
        throw;
    }
}

/** Safe insert to map-of-pointers. <tt>safeInsert(c, k, new Ele())</tt>
    is identical to <tt>c.insert(make_pair(k, new Ele()))</tt> but
    avoids the memory leak if insert throws.
    \return iterator pointing to newly-inserted element. If you
    use safeInsert(), it would probably be a leak if the element
    already exists before the call, therefore this does not return
    the usual pair<iterator,bool> returned by insert(). */
template<class Container, class Key, class Ele>
typename Container::iterator
safeInsert(Container& c, const Key& key, Ele* p)
{
    try {
        typedef typename Container::value_type vt;
        return c.insert(vt(key, p)).first;
    }
    catch (...) {
        delete p;
        throw;
    }
}

/** Delete all objects from a container-of-pointers. Strictly
    speaking, just delete'ing the elements is undefined behaviour, we
    have to set the elements to zero. */
template<class Container>
void
safeClear(Container& c)
{
    for (typename Container::iterator i = c.begin(); i != c.end(); ++i) {
        delete *i;
        *i = 0;
    }
}

/** Advance iterator by a number of steps. Avoids overruning the end.
    \return a+n */
template<typename Iterator>
Iterator
safeAdvance(Iterator a, Iterator end, uint_t n = 1)
{
    while (n > 0 && a != end) {
        ++a;
        --n;
    }
    return a;
}

#endif
